/*
 *  Referal whois client (RWhois)
 * 
 *	Written by:  Scott Williamson of Network Solutions Inc. 
 * 
 *	This code may be freely used. If bugs are found they
 *  should be sent to rwhois@internic.net.  This code can not
 *  be sold..
 *
 *	Tested on Sun4s running 4.1.*
 *
 *	To compile on BSD4.[23], SunOS4.1.*, SunOS3.x, and Ultrix:
 *
 *		cc -o rwhois rwhois.c
 *
 *	Usage: rwhois [-h hostname[:port] ] [ -a hostname[:port] ]  
 *			[-c config name ] [-s -m -v -d -q] question\n");
 *
 *
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <stdio.h>
#include <netdb.h>

/* #define	DEBUG 1 */


#define	SERVICE		"rwhois"
#define	PROTOCOL	"tcp"
#define	NORMPORT	43
#define MAXSERVERS	90
#define MAXBUFFER	128

#define DEFAULTSHOT 0
#define	SINGLESHOT	1
#define MULTISHOT	2

#define	TRAIL_RESET	0
#define	TRAIL_ADD	1
#define	TRAIL_COMP	2
#define	TRAIL_LIST	3

#define	REF_PRESET	"preset"
#define	REF_CMDLINE	"command line"
#define	REF_CONFFILE "config file"

#define	CONFPATH	"/usr/local/etc"

#define	VERSION	"B0.9.3"

#define	TRUE 	1
#define FALSE	0

#define	GOOD	0
#define	BAD		1

#define	NONE	0  /* no status printing */
#define	WISPER	1 /* bare minimum status printing */
#define	TALK	2 /* constant status update */
#define	SHOUT	3  /* detailed diagnostic update */

/* the possible host names for the rwhois server */
int current_referral;
struct SERVER {
	char *server;
	int port;
	char *referral;
	time_t	start_time; 	/* strat connection time */
	time_t	con_time;		/* connection time */
	time_t	ans_time;		/* answer time */
};

struct SERVER rwhois_servers[MAXSERVERS] =  { 
	{"slam.internic.net", 3636, "local"},  /* this is a test root server */ 
	{"rs.internic.net", 43, "local"}, /* traditional whois server */ 
	{ NULL, 0, NULL}	/* null pointers terminate the list */

};


#define PRESET_SERVERS 2

FILE *fin, *fout;
char *cmdname;
char *calloc();
int shots;  /* single or multi shot referral */
int verbose;	/* bug the user with info or not */
int referral = PRESET_SERVERS;
char *confdir;
time_t start_time, con_time, ans_time;


main(argc, argv)
int argc;
char *argv[];
{
	register int		i;
	int			c;
	char			*host;
	char retline[80];
	int finished = FALSE;
	char refhost[80];
	char refport[20];
	char cmdline[80];
	char altcmdline[80], *cmdptr;
	char tmpline[80];
	char server_name[80];
	char version[80];
	char location[80];
	int iscommand = 0, multiple = 0;
	char answer;
	int inseg = 0;
	char *ptr;
	int gotsomething = FALSE;
	int extreferral = FALSE;
	int change_server = FALSE;
	char *ptr_new_port, *index();
	int new_port_num;
	char *temp;
	char **quicklook;
	char *hold_referral;
	int cret;
	int start_over = FALSE;
	int extsee_also = FALSE;
	char *strstr();



	/* save the command name (usually called "rwhois") */
	cmdname = argv[0];

	argc--, argv++;

	printf("\nReferal Whois client version %s\n", VERSION);

	if( argc < 1 )
	{
		help();
		exit(-1);
	}

	shots = DEFAULTSHOT;  /* go through the entire refeal list */
	verbose = WISPER;	
	cmdptr = cmdline;


	server_trail( NULL, 0, NULL, NULL, TRAIL_RESET);


	/* look quickly thourgh the augv to see if things should be verbose */
	/* don't act on all options till defaults are setup                 */

	quicklook = argv;

	while( *quicklook)
	{
		if (!strcmp(*quicklook, "-v")) 
			verbose = TALK;
		else if (!strcmp(*quicklook, "-d"))
			verbose = SHOUT;
		else if(!strcmp(*quicklook, "-q"))
			verbose = NONE;
		++quicklook;
	}



	/* clear all unused server structs */

	for( i=referral; i < MAXSERVERS; ++i)
	{
		rwhois_servers[i].server = NULL;
		rwhois_servers[i].port = 0;
		rwhois_servers[i].referral = NULL;
	}


	/**************************************************************/
	/* grab the defaults frist then process the command line  ..  */
	/* command line overrides defaults                            */
	/* will get a chance to run a different config file to modigy */
	/* defaults again                                             */
	/**************************************************************/

	procrc( NULL);

	/* select options and build the command list for whois/rwhois server */

	tmpline[0] = '\0';
	while( *argv)
	{


		if (!strcmp(*argv, "-h")) 
		{


			if(!change_server)
			{
				change_server = TRUE;
				current_referral = 0;
				referral = 0;
			}

			++argv;

			fix_server( *argv, REF_CMDLINE);

		}
		else if (!strcmp(*argv, "-a")) 
		{
			++argv;

			fix_server( *argv, REF_CMDLINE);

		}
		else if (!strcmp(*argv, "-s")) 
			shots = SINGLESHOT;
		else if (!strcmp(*argv, "-m")) 
			shots = MULTISHOT;
		else if (!strcmp(*argv, "-v")) 
			verbose = TALK;
		else if (!strcmp(*argv, "-d")) 
			verbose = SHOUT;
		else if (!strcmp(*argv, "-q")) 
			verbose = NONE;
		else if (!strcmp(*argv, "-c")) 
		{
			temp = calloc( 1,(strlen( confdir) + strlen( *argv) + 1)); 
			sprintf(  temp, "%s%s", confdir, *(++argv));
			procrc( temp);
			free( temp);
		}
		else 
		{
		
			if( multiple )
			 	strcat( tmpline, " ");	
			strcat( tmpline, *argv);
			multiple = 1;
		}
		++argv;
	}


	strcpy( cmdline, tmpline);
	current_referral = 0;

	if( verbose > TALK )
		list_servers();

	if(contact( ))
		finished = TRUE;
	else if( verbose > WISPER)
	printf("\nConnnection made to %s port %d connection took %ld second%c\n",
			rwhois_servers[current_referral].server, 
			rwhois_servers[current_referral].port,
			(con_time - start_time), (con_time - start_time) > 1 ? 's' : ' ');


	server_trail( rwhois_servers[current_referral].server, 
			rwhois_servers[current_referral].port, 
			rwhois_servers[current_referral].referral,
			cmdptr, TRAIL_ADD);



	while( !finished )
	{

#ifdef DEBUG
printf("\nStarting loop [%s] current_referral = %d referral = %d",
cmdptr, current_referral, referral);
printf("\nfout = %ld fint %ld cmdptr = %s", *fout, *fin, cmdptr);
fflush(stdout);
list_servers();
#endif

	fprintf( fout, "%s\r\n", cmdptr);
	fflush( fout);

#ifdef DEBUG
printf("\nJust sent command line [%s]\n", cmdline);
fflush(stdout);
#endif

	cmdptr = cmdline;


	/* read back results */
	while (fgets(retline, 80,fin) != NULL)
	{

#ifdef DEBUG
printf("\n-%s", retline);
fflush(stdout);
#endif

		if(retline[0] == '%')
		{

			if(!strncmp( retline, "%RWhois", 7))
			{

				ptr = strstr( retline, "V-");
				if( sscanf( ptr, "%s %s", version, location))
				if( verbose )
					printf("\nRWhois Server at %s version %s\n", 
							location, version);
				fprintf( fout, "-RWhois V-%s\r\n", VERSION);
				fprintf( fout, "%s\r\n", cmdptr);
				fflush( fout);
			}


			/* receive referral  */
			if(sscanf( retline, "%%referral %s", refhost))
			{

				if( extreferral == FALSE )
				{
					referral = 0;
					current_referral = 0;
					start_over = TRUE;
				}

				hold_referral = calloc( 1,  
					(strlen( rwhois_servers[current_referral].server ) + 15));
				sprintf( hold_referral, "%s:%d", 
					rwhois_servers[current_referral].server, 
					rwhois_servers[current_referral].port);

#ifdef DEBUG
printf("\ngot referal [%s] -- [%s]\n", refhost, hold_referral);
fflush(stdout);
#endif
				if( verbose > WISPER )
					printf("\n[%d]-Received referral to %s\n",
						referral+1, refhost);

				if( referral < MAXSERVERS )
				{
					extreferral = TRUE;
#ifdef DEBUG
printf("\ngoing to fix_server with [%s] [%s]\n", refhost, hold_referral);
fflush(stdout);
#endif
					fix_server( refhost, hold_referral);
				}
				else
					printf("\nMust refuse referral to %s .. to many referrals",
						refhost);
			}
			else if(!strncmp( retline, "%see-also", 9 ))
			{


				ptr = index( retline, ':');
				ptr++;
				ptr = index( ptr, ':');
				*(ptr++) = '\0';

				strcpy( altcmdline, ptr);

				ptr = index( retline, ' ');
				strcpy( refhost, ++ptr);


				/* make sure that line is not term with return */
				if((ptr = index( altcmdline, '\n')))
					*ptr = NULL;

				++extsee_also;


#ifdef DEBUG
printf("\ngot see-also [%s] with query [%s]\n", refhost, altcmdline);
fflush(stdout);
#endif

				if( verbose > WISPER )
				{
					printf("\n[%d]-Received see-also to %s with query %s ",
						referral+1, refhost, altcmdline);
				}

				if( referral < MAXSERVERS )
					insert_server( refhost, current_referral+1);
				else
					printf("\nMust refuse see-also to %s .. to many referral",
						refhost);

				cmdptr = altcmdline;
			}
		}
		else 
		{
			printf("%s", retline);
			gotsomething = TRUE;
		}
	}


		if(!start_over ) 
			++ current_referral;

		start_over = FALSE;

#ifdef DEBUG
printf("\n before ++current_referral - [%d][%d][%d]\n", referral, 
	current_referral, finished);
fflush(stdout);
#endif

	if( !finished && referral && rwhois_servers[current_referral].server 
			&& (( shots == MULTISHOT) 
			|| ( extreferral && shots == DEFAULTSHOT) 
			|| ( extreferral && shots == SINGLESHOT && current_referral == 0)
			|| extsee_also ))
	{

		if( extsee_also )
			--extsee_also;
		/* close current file disc */
		fclose( fin); fclose(fout);
		fin = fout = 0;

		if( verbose > TALK)
			list_servers();

		/* try to contact the referral list */
		if(!(server_trail( rwhois_servers[current_referral].server, 
				rwhois_servers[current_referral].port, 
				rwhois_servers[current_referral].referral, 
				cmdptr, TRAIL_COMP)))
		{
			if( cret = contact( ))
				finished = TRUE;
			else if( verbose > WISPER)
			printf("\nConnnection made to %s port %d connection took %ld second%c\n",
					rwhois_servers[current_referral].server, 
					rwhois_servers[current_referral].port,
					(con_time - start_time), 
					(con_time - start_time) > 1 ? 's' : ' ');

			server_trail( rwhois_servers[current_referral].server, 
					rwhois_servers[current_referral].port, 
					rwhois_servers[current_referral].referral,
					cmdline, TRAIL_ADD);
		}
		else
				finished = TRUE;



	}
	else
		finished = TRUE;

		ans_time = time( &ans_time);

	if( verbose > TALK)
		printf("\nAnswer took %d second%c\n", ( ans_time - con_time),
			(ans_time - con_time) > 1 ? 's' : ' ');

	}

	if( verbose > WISPER)
		server_trail( NULL, 0, NULL, NULL, TRAIL_LIST);

	return 0;
}


contact( )
{

	int	socket_hand;
	int i = -1;
	int	c;
	char *tmp;
	struct sockaddr_in	sin;
	struct hostent		*host;
	struct servent		*service;
	char *itoa();
	int ret;
	char *ctime();

	/* run thru the list of whois servers till one is found
	 * that can be resoolved and connected to..
	 */

	while( rwhois_servers[current_referral].server && !fout)
	{


		ret = GOOD;

		start_time = time( &start_time);

		if( verbose )
			printf("\nTrying to connect to [%s] port [%d]\n", 
				rwhois_servers[current_referral].server, 
				rwhois_servers[current_referral].port);

		if (!(host = gethostbyname (rwhois_servers[current_referral].server)))
		{
			unsigned long	hostaddr;

			hostaddr = inet_addr (rwhois_servers[current_referral].server);
			if (!(host = gethostbyaddr (&hostaddr, 4, AF_INET)))
			{
				fprintf(stderr, "%s: can't find host '%s'.\n",
					cmdname, rwhois_servers[current_referral].server);
				++current_referral;
				ret = BAD;
			}

		}
		else
		{

			/* Create a socket */

			if ((socket_hand = socket(host->h_addrtype, SOCK_STREAM, 0, 0)) < 0) 
			{
				tmp = calloc( 1, (strlen( cmdname) + 10)); 
				sprintf (tmp, "%s: socket",MAXBUFFER , cmdname);
				perror(tmp);
				++current_referral;
				free( tmp);
				ret = BAD;
				continue;
			}

			/* Bind to that socket */
			sin.sin_family = host->h_addrtype;

			/* lookup the port number for the rwhois server */


			bcopy (host->h_addr, &sin.sin_addr, host->h_length);
			if(!rwhois_servers[current_referral].port)
			{
				if ((service = getservbyname(SERVICE, PROTOCOL)) == NULL) {
					fprintf(stderr, "%s: Don't know what service %s/%s is.\n",
					cmdname, SERVICE, PROTOCOL);
					fprintf(stderr, "Will try to use port %d\n", NORMPORT);
				}

				/* Connect to that port */
				if( service == NULL )
					sin.sin_port = NORMPORT ;
				else
					sin.sin_port = service->s_port;

			}
			else
				sin.sin_port = rwhois_servers[current_referral].port;


			if (connect (socket_hand, &sin, sizeof (sin), 0) < 0) 
			{
				tmp = calloc( 1, (strlen( cmdname) 
					+ strlen( rwhois_servers[current_referral].server) + 20)); 
				sprintf (tmp, "%s: <%s> on connect : ", cmdname, 
						rwhois_servers[current_referral].server);
				perror(tmp);
				++current_referral;
				free( tmp);
				ret = BAD;
			}
			else
			{


				con_time = time( &con_time);
				/* associate a STREAM to 'fin' and 'fout' from 'socket_hand' */
				/* if this does not work bad news .... so exit */

				if ((fin = fdopen(socket_hand, "r")) == NULL 
					|| (fout = fdopen(socket_hand, "w")) == NULL) {
					printf("\nProblem with fdopen for sockets ... sorry!\n");
					close(socket_hand);
					exit(1);
				}
			}
		}


	}


	if ( !fout )
		return( BAD);


	return( ret);

}




procrc(cfile)
char *cfile;
{
FILE *fp;
char rcfile[81];
char lrcfile[81];
char temp[81];
char cmd[10], act[80];
char logtmp[81];
char *ptr, *sname;
int ret;
int servnum = 0;
char *getenv();


	if( cfile)
	{
		if(verbose > WISPER)
			printf("\nTrying to override defaults with [%s] config file... \n",
					cfile);
		if( !(fp = fopen( cfile, "r")))
		{
			strcat( cfile, ".rc");
			if( !(fp = fopen( cfile, "r")))
			{
				if( verbose > TALK )
				printf("\nCould not open [%s][.rc] configuration file .... \n", 
					cfile);
				fflush(stdout);
			}
		}
	}

	if( !cfile || (cfile && !fp))
	{
		sprintf( lrcfile, "%s/.rwhois.rc", getenv( "HOME"));
		sprintf( rcfile, "%s/rwhois.rc", CONFPATH);


		if( !(fp = fopen( lrcfile, "r")))
		{
			if( !(fp = fopen( rcfile, "r")))
			{
				if( verbose > TALK )
				printf("\nNo [%s] or [%s] configuration file so ....  \
					\nusing defaults .... \n", rcfile, lrcfile);
				return(-1);
			}
		}
	}
		

	while( fscanf( fp, "%s %s", cmd, act) != EOF) 
	{
		upper( cmd);
		if((!strncmp( cmd, "ADDSERVER", 6)) && servnum < MAXSERVERS)
		{
			fix_server( act, REF_CONFFILE);
		}
		else if((!strncmp( cmd, "CHANGESERVER", 6)) && servnum < MAXSERVERS)
		{
			referral = 0;
			fix_server( act, REF_CONFFILE);
		}
		else if(!strncmp( cmd, "SHOTS", 5))
		{
			upper( act);
			if(!strncmp( act, "MULTI", 5))
			{
				if( verbose > WISPER )
					printf("\nSwitching to multi referral mode\n");
				shots = MULTISHOT;
			}
			else if(!strncmp( act, "SINGLE", 5))
			{
				if( verbose > WISPER )
					printf("\nSwitching to single referral mode\n");
				shots = SINGLESHOT;
			}
			else
			{
				printf("\nBad referal mode in config file must be \
						(single or multi )\n");
				shots = DEFAULTSHOT;
			}
		}
		else if(!strncmp( cmd, "VERBOSE", 6))
		{
			upper( act);
			if(!strncmp( act, "ON", 2))
				verbose = TALK;
			else if(!strncmp( act, "DEBUG", 5))
				verbose = SHOUT;
			else if(!strncmp( act, "QUITE", 5))
				verbose = NONE;
			else
				verbose = WISPER;
		}
		else if(!strncmp( cmd, "CONFDIR", 6))
		{
			confdir = calloc( 1,  strlen( act));
			strcpy( confdir, act);
			if( verbose > WISPER)
				printf("\nSeting the config directory to [%s]\n", confdir);
		}
		else
			printf("\nInvalid config file command [%s]\n", act);
	}

	fclose(fp);
	return(0);

}

help()
{
	printf("\n%s: %s [-h hostname[:port] ] [ -a hostname[:port] ] ", 
			cmdname, cmdname); printf("[-c config name ] \n\t\t[-s -m -v -q] question\n");
	printf("\n\t-h replace server list [ may be many ]");
	printf("\n\t-a add to server list [ may be many ]");
	printf("\n\t-c user config named [ may be many .. loaded in order ]");
	printf("\n\t\t[will try to add .rc to name if cant open]");
	printf("\n\t-s single server request .. no referral");
	printf("\n\t-m multiple referal");
	printf("\n\t-v verbose info to user");
	printf("\n\t-d debug ( very verbose )\n");
	printf("\n\t-q quite ( not verbose at all )\n");

}


upper( string )
char *string;
{
char *ptr;

        ptr = string;
        while(*ptr)
        {
                if((isalpha(*ptr)) && (islower(*ptr)))
                        *ptr = toupper(*ptr);
                ++ptr;
        }
}

fix_server( entry, refwhere)
char *entry;
char *refwhere;
{
char *ptr_new_port;
int new_port_num;
char *refstring;

#ifdef DEBUG
printf("\nfix_server %s -> [%s]", entry, refwhere);
fflush(stdout);
list_servers();
#endif

			if( ptr_new_port = index( entry, ':'))
			{
				*(ptr_new_port++) = NULL;
				new_port_num = atoi( ptr_new_port);
			}
			else
				new_port_num = NORMPORT; 

			if(!referral && verbose > WISPER )
				printf("Changing  whois server list to : [%s] port [%d]\n", 
					entry, new_port_num);
			else if( verbose > WISPER )
				printf("adding whois server to list : [%s] port [%d] \n", 
					entry, new_port_num);

			if( refwhere )
			{
				rwhois_servers[referral].referral = 
					calloc( 1,  (strlen( refwhere) + 1));
				strcpy( rwhois_servers[referral].referral ,  refwhere);
			}
			else
			{ 
				rwhois_servers[referral].referral = 
						calloc( 1,  10);
				strcpy( rwhois_servers[referral].referral ,  "not know");
			} 
				

			rwhois_servers[referral].server = calloc( 1, (strlen( entry)+1));
			strcpy( rwhois_servers[referral].server, entry);
			rwhois_servers[referral].port = new_port_num;

			++referral;

				/* make sure the referral list stops  */
			rwhois_servers[ referral].server = NULL; 
			rwhois_servers[ referral].port = 0; 
			rwhois_servers[ referral].referral = NULL;


}

insert_server( entry, where)
char *entry;
int where;
{
char *ptr_new_port;
int new_port_num;
int x;


			if( ptr_new_port = index( entry, ':'))
			{
				*(ptr_new_port++) = NULL;
				new_port_num = atoi( ptr_new_port);
			}
			else
				new_port_num = NORMPORT; 

			if((referral + 1) > MAXSERVERS )
				printf("\nCould not insert server %s port %d", 
					entry, new_port_num);
			if(verbose > TALK)
				printf("\nInserting whois server [%s] port [%d]\n", 
					entry, new_port_num);

			for( x = referral; x > where; --x)
			{
				rwhois_servers[x].server = rwhois_servers[x-1].server;
				rwhois_servers[x].port = rwhois_servers[x-1].port;
				rwhois_servers[x].referral = rwhois_servers[x-1].referral;
			}

			++referral;

			rwhois_servers[referral].server = NULL;
			rwhois_servers[referral].port = 0;
			rwhois_servers[referral].referral = NULL;


			rwhois_servers[where].server = calloc( 1, (strlen( entry)+1));
			strcpy( rwhois_servers[where].server, entry);
			rwhois_servers[where].port = new_port_num;
			rwhois_servers[where].referral = calloc( 1, (strlen( 
				rwhois_servers[current_referral].server)+8));
			sprintf( rwhois_servers[where].referral , "%s:%d",
					rwhois_servers[current_referral].server,
					rwhois_servers[current_referral].port);


}




/**************************************

	server_trail( server, port, query, action); 
	server = name of server
	port = port number on server
	query = question asked
	action = 0 = reset
			 1 = add trail
			 2 = compare to trail
			 3 = list trail
*****************************************/


server_trail( server, port, from, query, action)
char *server;
int port;
char *from;
char *query;
int action;
{
static char *server_list[MAXSERVERS];
static int hop_count;
int x, ret = 0;
char *comp_str;

#ifdef DEBUG
printf("\nInside server_trail [%s] [%d] [%s] [%s] [%d]\n",
server, port, from, query, action);
fflush(stdout);
#endif

	switch( action )
	{
		case TRAIL_RESET:
			hop_count = 0;
			break;

		case TRAIL_ADD:

				if( server)
				{
					if( hop_count > MAXSERVERS )
					{
						if( verbose  > WISPER )

						printf("\nExhausted server trail list .... skipping\n");
						break;
					}
				/*	upper( query);  */
					server_list[hop_count]  = calloc( 1,     
						(strlen( server) + 8 + strlen( from) 
							+ strlen( query) + 5));
					sprintf( server_list[hop_count++], "%s:%d->%s[%s]", 
							server, port, from, query);
				}
				break;

		case TRAIL_LIST:
			printf("\nRWhois server trail: [ %d hops ]\n", hop_count);
				for( x = 0; x < hop_count; ++x)
					printf("\n[%d] - %s", x+1, server_list[x]);
			putchar('\n');
			break;

		case TRAIL_COMP:
			comp_str  = calloc( 1, (strlen( server) + 15 + strlen( from) 
					+ strlen( query)));
			sprintf( comp_str, "%s:%d->%s[%s]", server, port, from, query);
			for( x = 0; x < hop_count && !ret; ++x)
			{
#ifdef DEBUG
printf("\nMakeing referral compare \n[%s]\n[%s]\n", comp_str, server_list[x]);
fflush(stdout);
#endif
				if(!strcmp( comp_str, server_list[x]))
				{
			printf("\nFound recursive server call ... stoping this madness!!\n");
					ret = -1;
				}
			}
			free( comp_str);
			break;


	}

	return( ret);

}

list_servers()
{
int x;
int current;

	if( current_referral < 0 )
		current = 0;
	else
		current = current_referral;

	printf("\nCurrent server list [%d:%d]:\n", referral, current_referral);
	for( x = current_referral; x < referral; ++x)
		printf("%s port %d -> %s\n", rwhois_servers[x].server,
			rwhois_servers[x].port, rwhois_servers[x].referral);

}
		
